create table countries (
    "areaInSqKm" real,
    capital text,
    continent text,
    "continentName" text,
    "countryCode" text,
    "countryName" text,
    "currencyCode" text,
    east real,
    "fipsCode" text,
    "isoAlpha3" text,
    "isoNumeric" text,
    languages text,
    north real,
    population integer,
    south real,
    west real
);
